"""Random selectors."""


from zeph.selectors.abstract import AbstractSelector
from zeph.typing import Network


class RandomSelector(AbstractSelector):
    def select(self, agent_uuid: str) -> set[Network]:
        return self._select_random(agent_uuid)



from collections import OrderedDict, defaultdict

from zeph.rankers import AbstractRanker
from zeph.typing import Agent, Link, Network


class NaiveRanker(AbstractRanker):
    def __call__(
        self, links: dict[tuple[Agent, Network], set[Link]]
    ) -> dict[Agent, list[Network]]:
        all_links: set[Link] = set()
        covered: set[Link] = set()
        prefixes: dict[Agent, list[Network]] = defaultdict(list)

        for links_ in links.values():
            all_links.update(links_)

        # Sort the subsets by size in descending order
        subsets = OrderedDict(
            sorted(
                [(k, v) for k, v in links.items()],
                key=lambda x: len(x[1]),
                reverse=True,
            )
        )

        for agent, prefix in subsets:
            if covered == all_links:
                break
            if subsets[(agent, prefix)] - covered:
                prefixes[agent].append(prefix)
                covered.update(subsets[(agent, prefix)])

        return prefixes



"""
Constrained selectors.

Add the constraint of a prefix being probed by exactly one agent.
It's less effective than letting the agents potentially probe the same prefixes (see paper).
"""

from collections import defaultdict
from itertools import cycle

from zeph.selectors.random import RandomSelector
from zeph.typing import Network


class ConstrainedRandomSelector(RandomSelector):
    def __init__(self, universe: set[Network], budgets: dict[str, int]) -> None:
        super().__init__(universe, budgets)
        self.prefixes = self.dispatch()

    def dispatch(self) -> dict[str, set[Network]]:
        prefixes: dict[str, set[Network]] = defaultdict(set)
        for agent, prefix in zip(cycle(self.budgets), self.universe_shuffled()):
            if len(prefixes[agent]) >= self.budgets[agent]:
                continue
            prefixes[agent].add(prefix)
        return prefixes

    def select(self, agent_uuid: str) -> set[Network]:
        return self.prefixes[agent_uuid]



"""
Zeph.

Communicate with Iris to perform measurements.
"""
import logging
from pathlib import Path
from typing import Optional

import typer
from iris_client import IrisClient
from pych_client import ClickHouseClient
from tqdm import tqdm

from zeph import rankers
from zeph.iris import (
    create_measurement,
    get_agents,
    get_measurement_agents,
    upload_prefix_list,
)
from zeph.logging import logger
from zeph.queries import GetUniqueLinksByPrefix
from zeph.rankers import AbstractRanker
from zeph.selectors import EpsilonSelector
from zeph.typing import Network

app = typer.Typer()


@app.command()
def zeph(
    prefixes_file: Path = typer.Argument(
        ...,
        help="File containing all the /24 or /64 prefixes that can be probed.",
    ),
    previous_uuid: Optional[str] = typer.Argument(
        None,
        help="UUID of the previous measurement cycle",
    ),
    agent_tag: str = typer.Option(
        "all",
        help="The tag used to select the agents",
        metavar="TAG",
    ),
    measurement_tags: str = typer.Option(
        "test,zeph",
        help="The tags that will be associated to the measurement",
        metavar="TAGS",
    ),
    tool: str = typer.Option(
        "diamond-miner",
        help="The measurement tool",
        metavar="TOOL",
    ),
    protocol: str = typer.Option(
        "icmp",
        help="The probe protocol",
        metavar="PROTOCOL",
    ),
    min_ttl: int = typer.Option(
        2,
        help="The minimum probe TTL",
        metavar="TTL",
    ),
    max_ttl: int = typer.Option(
        32,
        help="The maximum probe TTL",
        metavar="TTL",
    ),
    exploration_ratio: float = typer.Option(
        0.1,
        help="The minimum percentage of the budget allocated to exploration",
    ),
    fixed_budget: Optional[int] = typer.Option(
        None,
        help="Override the agents budget",
    ),
    ranker_class: str = typer.Option(
        "DFGCoverRanker",
        help="The class to use to rank prefixes",
    ),
    dry_run: bool = typer.Option(
        False,
        help="Do not actually perform the measurement",
    ),
    iris_base_url: str = typer.Option(
        None,
        help="Iris API URL",
        metavar="BASE_URL",
    ),
    iris_username: str = typer.Option(
        None,
        help="Iris API username",
        metavar="USERNAME",
    ),
    iris_password: str = typer.Option(
        None,
        help="Iris API password",
        metavar="PASSWORD",
    ),
) -> None:
    logging.basicConfig(level=logging.INFO)
    universe = set()
    with prefixes_file.open() as f:
        for line in tqdm(f):
            if line.startswith("#"):
                continue
            line = line.strip()
            assert line.endswith("/24") or line.endswith("/64")
            universe.add(line)
    logger.info("file=%s distinct-prefixes=%s", prefixes_file, len(universe))

    with IrisClient(
        base_url=iris_base_url,
        username=iris_username,
        password=iris_password,
    ) as iris:
        params = {}
        if previous_uuid:
            params["measurement_uuid"] = previous_uuid
        credentials = iris.get("/users/me/services", params=params).json()
        with ClickHouseClient(**credentials["clickhouse"]) as clickhouse:
            run_zeph(
                iris=iris,
                clickhouse=clickhouse,
                ranker=ranker_class,
                universe=universe,
                agent_tag=agent_tag,
                measurement_tags=measurement_tags.split(","),
                tool=tool,
                protocol=protocol,
                min_ttl=min_ttl,
                max_ttl=max_ttl,
                exploration_ratio=exploration_ratio,
                previous_uuid=previous_uuid,
                fixed_budget=fixed_budget,
                dry_run=dry_run,
            )


def run_zeph(
    *,
    iris: IrisClient,
    clickhouse: ClickHouseClient,
    ranker: AbstractRanker | str,
    universe: set[Network],
    agent_tag: str,
    measurement_tags: list[str],
    tool: str,
    protocol: str,
    min_ttl: int,
    max_ttl: int,
    exploration_ratio: float,
    previous_uuid: str | None,
    fixed_budget: int | None,
    dry_run: bool,
) -> None:
    if isinstance(ranker, str):
        ranker_ = getattr(rankers, ranker)()
    else:
        ranker_ = ranker
    # Rank the prefixes based on the previous measurement
    ranked_prefixes = {}
    if previous_uuid:
        logger.info("get-previous-agents")
        previous_agents = get_measurement_agents(iris, previous_uuid)
        logger.info("previous-agents=%s", previous_agents)

        logger.info("get-previous-links")
        query = GetUniqueLinksByPrefix(filter_virtual=True)
        links = query.for_all_agents(clickhouse, previous_uuid, previous_agents)

        logger.info("rank-previous-prefixes")
        ranked_prefixes = ranker_(links)

    logger.info("get-current-agents")
    agents = get_agents(iris, agent_tag)
    logger.info("current-agents=%s", list(agents.keys()))

    logger.info("compute-budget")
    budgets: dict[str, int] = {}
    for agent_uuid, agent in agents.items():
        if fixed_budget:
            budgets[agent_uuid] = fixed_budget
        else:
            # Compute the budget (number of prefixes to send per agent).
            # Based on the probing rate and the approximate duration of the measurement.
            # 6 hours at 100'000 pps -> 200'000 prefixes (from the paper)
            probing_rate = agent["parameters"]["max_probing_rate"]
            budgets[agent_uuid] = probing_rate * 2
        logger.info("agent=%s budget=%s", agent_uuid, budgets[agent_uuid])

    # Instantiate the selector
    selector = EpsilonSelector(universe, budgets, exploration_ratio, ranked_prefixes)

    # Select and upload the prefixes
    targets = {}
    for agent_uuid in agents:
        logger.info("agent=%s select-prefixes", agent_uuid)
        prefixes = selector.select(agent_uuid)
        if not dry_run:
            logger.info("agent=%s upload-prefixes", agent_uuid)
            targets[agent_uuid] = upload_prefix_list(
                iris, prefixes, protocol, min_ttl, max_ttl
            )
            logger.info("agent=%s key=%s", agent_uuid, targets[agent_uuid])

    # Create the measurement
    if not dry_run:
        logger.info("create-measurement")
        definition = {
            "tags": measurement_tags,
            "tool": tool,
            "agents": [
                {
                    "uuid": agent_uuid,
                    "target_file": targets[agent_uuid],
                    "tool_parameters": {
                        "flow_mapper": "RandomFlowMapper",
                        "flow_mapper_kwargs": {"seed": 2021},
                    },
                }
                for agent_uuid in agents
            ],
        }
        measurement = create_measurement(iris, definition)
        logger.info("measurement_uuid=%s", measurement["uuid"])



from collections import defaultdict

from zeph.rankers import AbstractRanker
from zeph.typing import Agent, Link, Network


class GreedyCoverRanker(AbstractRanker):
    def __call__(
        self, links: dict[tuple[Agent, Network], set[Link]]
    ) -> dict[Agent, list[Network]]:
        all_links: set[Link] = set()
        covered: set[Link] = set()
        prefixes: dict[Agent, list[Network]] = defaultdict(list)

        for links_ in links.values():
            all_links.update(links_)

        while covered != all_links and links:
            agent, prefix = max(links, key=lambda k: len(links[k] - covered))
            prefixes[agent].append(prefix)
            covered.update(links[(agent, prefix)])
            links.pop((agent, prefix))

        return prefixes



"""
Epsilon-based selectors.

Apply a reinforcement-learning approach to select prefixes.
"""


from zeph.selectors.abstract import AbstractSelector
from zeph.typing import Network


class EpsilonSelector(AbstractSelector):
    def __init__(
        self,
        universe: set[Network],
        budgets: dict[str, int],
        epsilon: float,
        ranked_prefixes: dict[str, list[Network]],
    ):
        super().__init__(universe, budgets)
        self.epsilon = epsilon
        self.ranked_prefixes = ranked_prefixes

    def select(self, agent_uuid: str) -> set[Network]:
        """
        epsilon-based policy :
            * select e where eB will be used for exploration.
              and (1 - e)B is used for exploitation
            * Get the (1-e)B previous prefixes that maximize the links
            * Pick random eB prefixes not already used in the exploration set
        """
        # Compute the number of prefixes for exploration [eB] / exploitation [(1-e)B]
        n_prefixes_exploitation = int((1 - self.epsilon) * self.budgets[agent_uuid])

        # Pick the (1-e)B prefix with the best reward
        rank = self.ranked_prefixes.get(agent_uuid, [])
        prefixes = set(rank[:n_prefixes_exploitation])

        # Add random prefixes until the budget is completely burned
        return self._select_random(agent_uuid, preset=prefixes)



from collections import Counter, defaultdict

from zeph.rankers import AbstractRanker
from zeph.typing import Agent, Link, Network


class UniqueLinksRanker(AbstractRanker):
    def __call__(
        self, links: dict[tuple[Agent, Network], set[Link]]
    ) -> dict[Agent, list[Network]]:
        """
        links: (agent, prefix) -> links
        """
        # Count how many times each link has been seen
        counts: Counter[Link] = Counter()
        for links_ in links.values():
            counts.update(links_)

        # Compute the reward as the number of globally unique (across all agents) links a prefix has seen.
        rewards: dict[Agent, dict[Network, int]] = defaultdict(lambda: defaultdict(int))
        for (agent, prefix), links_ in links.items():
            for link in links_:
                if counts[link] == 1:
                    rewards[agent][prefix] += 1

        # Sort the prefixes by reward
        prefixes: dict[Agent, list[Network]] = {}
        for agent, rewards_ in rewards.items():
            prefixes[agent] = [
                x[0] for x in sorted(rewards_.items(), key=lambda x: x[1])
            ]

        return prefixes
